from fenics import *
import numpy as np
import pickle
import sympy


def get_locally_stable_fp(A, H):
    x = sympy.Symbol('x')
    a = sympy.Symbol('a')
    h = sympy.Symbol('h')
    solution = sympy.solve(4*a*x*(x**2 - 1) - h, x)

    #need to take real part, because sympy will give very small imaginary part otherwise
    numerical_solution = [sympy.N(sympy.re(solution[i].subs([(a,A),(h,H)]))) for i in range(3)]

    return [float(min(numerical_solution)), float(max(numerical_solution))]


class InitialCondition(UserExpression):
    def __init__(self, x_threshold, phi_plus, phi_minus):
        super().__init__()
        self.x_threshold = x_threshold
        self.phi_plus = phi_plus
        self.phi_minus = phi_minus

    def eval_cell(self, value, x, ufc_cell):
        if (x[0] <= self.x_threshold):
            value[0] = self.phi_plus
        else:
            value[0] = self.phi_minus

""" This gives a trejectory of an Allen-Cahn system in 1D

Args:
    H (float): field strength
    A (float): strength of double-well potential
    kappa (float): Diffusivity
    num_steps (int): number of simulated time steps
    step_length (float): length of time step
    
Returns:
    time_series (list): a list of numpy arrays which contain the field's values at the nodes of the mesh
"""
def get_trajectory(H, A=1000., kappa = 1., num_steps = 5000, step_length = 0.0001, system_length = 1.):
    #set up mesh
    num_elements = 200
    mesh = IntervalMesh(num_elements, 0, system_length)

    # set up function space
    # takes arguments as mesh, type of element,
    # degree of element, specify PBC as constrained region
    V = FunctionSpace(mesh, 'P', 1)
    # define trial function and function
    u = Function(V)
    v = TestFunction(V)
    u_n = Function(V)

    #time discretization
    dt = Constant(step_length)


    # Define initial condition
    phi_minus, phi_plus = get_locally_stable_fp(A,H)
    initial_condition = InitialCondition(.3, phi_plus, phi_minus)
    u_n.interpolate(initial_condition)

    # prepare constants for fenics simulation
    H_const = Constant(H)
    A_const = Constant(4*A)
    Kappa_const = Constant(kappa)
    one = Constant(1.)

    # define RD-equation
    F = (u - u_n)*v*dx + dt*Kappa_const*dot(grad(u), grad(v))*dx + dt*(A_const*u*(u*u - one) -H_const)*v*dx

    # initialize time series
    t = 0
    time_series = []

    #turn off log, otherwise there's a lot of text output
    set_log_level(40)

    # perform time steps
    for step in range(num_steps):
        t += dt
        if(step%1000==0):
            print("in time step " + str(step))

        solve(F ==0, u)
        u_n.assign(u)
        if step%10 == 0:
            new_state = np.asarray(u.compute_vertex_values(mesh))
            time_series.append(new_state)

    return time_series

if __name__=="__main__":
    time_series = get_trajectory(10.)

    with open("test_storage1DReflective.pickle", 'wb') as handle:
        pickle.dump(time_series, handle, protocol=pickle.HIGHEST_PROTOCOL)

